(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RedPairImpl imports PolyOrder Matrix RPO SCT Rational
begin

datatype arctic = MinInfty | Num_arc int

types 'f lpo_repr = "(('f \<times> nat) \<times> nat)list"

datatype 'f redpair = Nat_carrier "('f,nat)lpoly_interL" 
  | Int_carrier "('f,int)lpoly_interL" 
  | Arctic_carrier "('f,arctic)lpoly_interL" 
  | Int_mat_carrier nat "('f,int mat)lpoly_interL"
  | Arc_mat_carrier nat "('f,arctic mat)lpoly_interL"
  | LPO "'f lpo_repr" "'f afs_list"
  | Rat_carrier rat "('f,rat)lpoly_interL"
  | Rat_mat_carrier rat nat "('f,rat mat)lpoly_interL"

fun plus_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> arctic"
where "plus_arctic MinInfty y = y"
   |  "plus_arctic x MinInfty = x"
   |  "plus_arctic (Num_arc x) (Num_arc y) = (Num_arc (max x y))"

fun mult_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> arctic"
where "mult_arctic MinInfty y = MinInfty"
   |  "mult_arctic x MinInfty = MinInfty"
   |  "mult_arctic (Num_arc x) (Num_arc y) = (Num_arc (x + y))"

fun geq_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> bool"
where "geq_arctic x MinInfty = True"
   |  "geq_arctic MinInfty (Num_arc _) = False"
   |  "geq_arctic (Num_arc x) (Num_arc y) = (x \<ge> y)"

fun gt_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> bool"
where "gt_arctic _ MinInfty = True"
   |  "gt_arctic MinInfty (Num_arc _) = False"
   |  "gt_arctic (Num_arc x) (Num_arc y) = (x > y)"

instantiation arctic :: "shows" begin
  fun shows_arctic :: "arctic \<Rightarrow> shows"
  where "shows_arctic(Num_arc i) = shows i"
      | "shows_arctic(MinInfty)  = shows_string ''-inf''"
  definition shows_list_arctic where "shows_list_arctic(as::arctic list) \<equiv> shows_list_aux shows as"

  lemma assoc_arctic: "shows (a::arctic) r @ s = shows a (r@s)" by (cases a) auto
  lemma assoc_arctic_elt: "\<forall>(a::arctic)\<in>set as. shows a r @ s = shows a (r@s)"
    using assoc_arctic by auto
  instance by
    (intro_classes,rule assoc_arctic,unfold shows_list_arctic_def)
    (rule shows_list_aux_assoc[OF assoc_arctic_elt])
end

definition nat_name :: "string" where "nat_name = ''naturals''"
definition int_name :: "string" where "int_name = ''naturals with negative constants''"
definition rat_name :: "string" where "rat_name = ''rationals (with negative constants)''"
definition int_mat_name :: "string"
where "int_mat_name = ''matrices over the naturals with negative constants''"
definition rat_mat_name :: "string"
where "rat_mat_name = ''matrices over the rationals (with negative constants)''"
definition arc_mat_name :: "string" where "arc_mat_name = ''matrices over the arctic semi-ring''"
definition arctic_name :: "string" where "arctic_name = ''arctic semiring''"


lemma int_SN : "SN {(x :: int,y). 0 \<le> y \<and> x > y}"
proof -
  have "{(x,y). 0 \<le> x \<and> (x :: int) < y} = ({(x,y). y \<ge> 0 \<and> (x :: int) > y}^-1)" by auto
  with wf_int_ge_less_than[where d = 0] have wf: "wf ({(x,y). y \<ge> 0 \<and> (x :: int) > y}^-1)" unfolding int_ge_less_than_def 
    by simp
  show "SN {(x :: int,y). 0 \<le> y  \<and> x > y}"
    by (rule wf_imp_SN, rule wf)
qed

interpretation nat_poly_carrier: poly_carrier "\<lambda> x. True" "0 :: nat" "op +" "op *" "op \<ge>" "op >" True
  by (unfold_locales, auto simp: nat_distrib)


(* the minimum of \<delta> and 1 is only taken to ensure that 1 > 0 *)
definition rat_gt :: "rat \<Rightarrow> rat \<Rightarrow> rat \<Rightarrow> bool" where 
  "rat_gt \<delta> x y \<equiv> x - y \<ge> (if \<delta> > 0 then min \<delta> 1 else 1)"


interpretation rat_poly_carrier: poly_carrier "\<lambda> x. True" "0 :: rat" "op +" "op *" "op \<ge>" "rat_gt \<delta>" True
  by (unfold_locales, auto simp: rat_gt_def Ring_and_Field.mult_right_mono Ring_and_Field.mult_left_mono Ring_and_Field.right_distrib Ring_and_Field.left_distrib)

interpretation int_poly_carrier: poly_carrier "\<lambda> x. True" "0 :: int" "op +" "op *" "op \<ge>" "op >" True
  by (unfold_locales, auto simp: int_distrib Ring_and_Field.mult_right_mono Ring_and_Field.mult_left_mono)

interpretation arctic_poly_carrier: poly_carrier "\<lambda> x. True" MinInfty plus_arctic mult_arctic geq_arctic gt_arctic False
proof (unfold_locales, auto)
  fix x y
  show "plus_arctic x y = plus_arctic y x"
    by (cases x, cases y, auto, cases y, auto)
next
  fix x y z
  show "plus_arctic x (plus_arctic y z) = plus_arctic (plus_arctic x y) z"
    by (cases x, auto, cases y, auto, cases z, auto)
next
  fix x y z
  show "mult_arctic x (mult_arctic y z) = mult_arctic (mult_arctic x y) z"
    by (cases x, auto, cases y, auto, cases z, auto)
next
  fix x 
  show "mult_arctic x MinInfty = MinInfty"
    by (cases x, auto)
next
  fix x y z
  show "mult_arctic x (plus_arctic y z) = plus_arctic (mult_arctic x y) (mult_arctic x z)"
    by (cases x, auto, cases y, auto, cases z, auto)
next
  fix x y z
  show "mult_arctic (plus_arctic x y) z = plus_arctic (mult_arctic x z) (mult_arctic y z)"
    by (cases x, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z
  assume "geq_arctic x y" and "gt_arctic y z"
  thus "gt_arctic x z"
    by (cases z, simp, cases y, simp, cases x, auto)
next 
  fix x y z
  assume "gt_arctic x y" and "geq_arctic y z"
  thus "gt_arctic x z"
    by (cases z, simp, cases y, simp, cases x, auto)
next
  fix x y z
  assume "geq_arctic x y" 
  thus "geq_arctic (plus_arctic x z) (plus_arctic y z)"
    by (cases x, cases y, cases z, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z
  assume "geq_arctic x y" 
  thus "geq_arctic (mult_arctic x z) (mult_arctic y z)"
    by (cases x, cases y, cases z, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z
  assume "geq_arctic y z" 
  thus "geq_arctic (mult_arctic x y) (mult_arctic x z)"
    by (cases x, cases y, cases z, auto, cases y, cases z, auto, cases z, auto)
next
  fix x
  show "geq_arctic x x" 
    by (cases x, auto)
next
  fix x y z
  assume "geq_arctic x y" and "geq_arctic y z"
  thus "geq_arctic x z"
    by (cases x, cases y, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z u
  assume "gt_arctic x y" and "gt_arctic z u"
  thus "gt_arctic (plus_arctic x z) (plus_arctic y u)"
    by (cases y, cases u, simp, cases z, auto, cases x, auto, cases u, auto, cases z, auto, cases x, auto, cases x, auto, cases z, auto, cases x, auto)
next
  fix x y z
  assume "gt_arctic x y"
  thus "gt_arctic (mult_arctic x z) (mult_arctic y z)"
    by (cases y, simp, cases z, simp, cases x, auto)
next
  fix x
  assume "gt_arctic MinInfty x"
  thus "x = MinInfty"
    by (cases x, auto)
qed


interpretation nat_poly_order_carrier : poly_order_carrier "\<lambda> x. True" "0 :: nat" "op +" "op *" "op \<ge>" "op >" True 1 id nat_name "\<lambda> x. True"
proof (unfold_locales, auto)
  have "{(x,y). (x :: nat) < y} = ({(x,y). (x :: nat) > y}^-1)" by auto
  with wf_less have wf: "wf ({(x,y). (x :: nat) > y}^-1)" by simp 
  show "SN {(x :: nat,y). x > y}" 
    by (rule wf_imp_SN, rule wf)
qed

context poly_order_carrier
begin
declare arc_pos_max0[simp del]
end

interpretation int_poly_order_carrier : poly_order_carrier "\<lambda> x. True" "0 :: int" "op +" "op *" "op \<ge>" "op >" True 1 "max 0" int_name "\<lambda> x. True"
  by (unfold_locales, auto, rule int_SN)

lemma rat_gt_SN : "SN {(x,y). 0 \<le> y \<and> rat_gt \<delta> x y}"
proof (rule ccontr, unfold SN_defs, clarify)
  fix x f
  let ?R = "{(x,y). 0 \<le> y \<and> rat_gt \<delta> x y}"
  let ?S = "\<lambda> d. {(x,y). (0 :: rat) \<le> y \<and> x - y \<ge> d}" 
  assume Rsteps: "\<forall> i. (f i, f (Suc i)) \<in> ?R"
  have "\<exists> d. 0 < d \<and> ?R = ?S d"
    by (cases "\<delta> > 0", simp add: rat_gt_def, rule exI[of _ "min \<delta> 1"], simp, simp add: rat_gt_def, rule exI[of _ 1], simp)
  then obtain d :: rat where d: "d > 0" and R: "?R = ?S d" by auto
  let ?Rd = "?S d"
  from Rsteps R have steps:  "\<forall> i. (f i, f (Suc i)) \<in> ?Rd" by auto
  from spec[OF steps, of 0] d have fz_0: "f 0 > 0" by auto
  {
    fix n :: nat
    have "f n \<le> f 0 - (of_nat n) * d"
    proof (induct n, simp)
      case (Suc n)
      with spec[OF steps, of n] show ?case by auto
    qed
  } note main1 = this
  { 
    fix n :: nat
    have "f 0 \<ge> of_nat (Suc n) * d" using main1[of "Suc n"] spec[OF steps, of "n"] by auto
  } note main2 = this
  obtain numeratord denomd where d_id: "d = Fract numeratord denomd" by (cases "d", auto)
  with d have denomd: "denomd \<noteq> 0" by (cases "denomd = 0", simp_all add: rat_number_collapse)
  obtain numerator0 denom0 where fz: "f 0 = Fract numerator0 denom0" by (cases "f 0", auto)
  with fz_0 have denom0: "denom0 \<noteq> 0" by (cases "denom0 = 0", simp_all add: rat_number_collapse)
  from fz d_id main2 have contra: "\<And> n. Fract numerator0 denom0 \<ge> Fract (int (Suc n) * numeratord) denomd" by (simp add: of_nat_rat)
  from denom0 have square: "denom0 * denom0 > 0" by  (cases "denom0 > 0", simp add: Ring_and_Field.ordered_semiring_strict_class.mult_pos_pos, simp add: Ring_and_Field.ordered_ring_strict_class.mult_neg_neg)
  from d d_id denomd have pos: "numeratord * denomd  > 0" 
    by (auto simp: sgn_times mult_less_0_iff zero_less_mult_iff less_rat' zero_less_Fract_iff, cases "numeratord = 0", auto simp: rat_number_collapse, cases "denomd < 0", simp_all,
      simp only: minus_rat_cancel[symmetric, of numeratord denomd],
      simp only: zero_less_Fract_iff[of "- denomd" "- numeratord"])
  let ?p = "numeratord * denom0 * denomd * denom0"
  have id: "?p = (denom0 * denom0) * (numeratord * denomd)" by auto
  have posp: "?p > 0" using Ring_and_Field.ordered_semiring_strict_class.mult_pos_pos[OF square pos] by (simp add: id)
  have "\<exists> p . ?p = p + 1 \<and> p \<ge> 0"
    by (rule exI[of _ "?p - 1"], simp add: posp)
  from this obtain p where p: "?p = p + 1" and ppos: "p \<ge> 0" by auto
  have id2: "\<And> n. n * numeratord * denom0 * denomd * denom0 = n * ?p" by auto
  from contra square denomd denom0 show False 
  proof (simp add: p id2)
    let ?r = "numerator0 * denomd * denomd * denom0"
    let ?rr = "?r - p - 1"
    assume ass: "\<And> n. p + 1 + int n * p + int n \<le> ?r"
    have nppos: "\<And> n. int n * p \<ge> 0" by  (rule Ring_and_Field.pordered_cancel_semiring_class.mult_nonneg_nonneg[OF _ ppos], simp)
    {
      fix n
      have "int n = 0 + int n" by auto
      also have "\<dots> \<le> int n * p + int n" using nppos[of n] by auto
      finally have "p + 1 + int n \<le> p + 1 + int n * p + int n" by auto
      with ass[of n] have "int n \<le> ?rr" by auto
    } note contra2 = this
    show False 
    proof (cases "?rr \<ge> 0")
      case False
      with contra2[of 0] show False by auto
    next
      case True
      from zero_le_imp_eq_int[OF True] obtain n where id2: "?rr = int n" by auto
      from contra2[of "Suc n"] show False by (simp only: id2)
    qed
  qed
qed

interpretation rat_poly_order_carrier : poly_order_carrier "\<lambda> x. True" "0 :: rat" "op +" "op *" "op \<ge>" "rat_gt \<delta>" True 1 "max 0" rat_name "\<lambda> x. True"
  by (unfold_locales, auto simp: rat_gt_SN, auto simp: rat_gt_def)

fun get_arctic_num :: "arctic \<Rightarrow> int"
where "get_arctic_num (Num_arc n) = n"

fun pos_arctic :: "arctic \<Rightarrow> bool"
where "pos_arctic MinInfty = False"
    | "pos_arctic (Num_arc n) = (n \<ge> 0)"

interpretation arctic_poly_order_carrier : poly_order_carrier "\<lambda> x. True" MinInfty plus_arctic mult_arctic geq_arctic gt_arctic False "Num_arc 0" id arctic_name pos_arctic 
proof (unfold_locales, auto)
  fix x 
  show "mult_arctic (Num_arc 0) x = x" by (cases x, auto)
next
  fix x 
  show "mult_arctic x (Num_arc 0) = x" by (cases x, auto)
next
  show "SN {(x,y). pos_arctic y \<and> gt_arctic x y}" 
    unfolding SN_defs
  proof (rule allI, rule notI)
    fix x
    assume "\<exists> f . f 0 = x \<and> (\<forall> i. (f i, f (Suc i)) \<in> {(x,y). pos_arctic y \<and> gt_arctic x y})"
    from this obtain f where "f 0 = x" and seq: "\<forall> i. (f i, f (Suc i)) \<in> {(x,y). pos_arctic y \<and> gt_arctic x y}" by auto
    from seq have steps: "\<forall> i. gt_arctic (f i) (f (Suc i)) \<and> pos_arctic (f (Suc i)) " by auto
    let ?g = "\<lambda> i. get_arctic_num (f i)"
    have "\<forall> i. ?g (Suc i) \<ge> 0 \<and> ?g i  > ?g (Suc i)"
    proof
      fix i
      from steps have i: "gt_arctic (f i) (f (Suc i)) \<and> pos_arctic (f (Suc i))" by auto
      from i obtain n where fi: "f i = Num_arc n" by (cases "f (Suc i)", simp, cases "f i", auto)
      from i obtain m where fsi: "f (Suc i) = Num_arc m" by (cases "f (Suc i)", auto)
      with i have gz: "m \<ge> 0" by simp
      from i fi fsi have "n > m" by auto
      with fi fsi gz
      show "?g (Suc i) \<ge> 0 \<and> ?g i > ?g (Suc i)" by auto
    qed
    from this obtain g where "\<forall> i. g (Suc i) \<ge> 0 \<and> (op > :: int \<Rightarrow> int \<Rightarrow> bool) (g i) (g (Suc i))" by auto
    hence "\<exists> f. f 0 = g 0 \<and> (\<forall> i. (f i, f (Suc i)) \<in> {(x,y). y \<ge> 0 \<and> x > y})" by auto
    with int_SN show False unfolding SN_defs by blast
  qed 
next
  fix x y
  assume "pos_arctic x" 
  thus "pos_arctic (plus_arctic x y)" by (cases x, simp, cases y, auto)
next
  fix x y
  assume "pos_arctic x" and "pos_arctic y"
  thus "pos_arctic (mult_arctic x y)" by (cases x, simp, cases y, auto)
qed

interpretation nat_mono_poly_order_carrier : mono_poly_order_carrier "\<lambda> x. True" "0 :: nat" "op +" "op *" "op \<ge>" "op >" True 1 id nat_name "\<lambda> x. True" "\<lambda> x. x \<noteq> 0"
  by (unfold_locales, auto)

interpretation int_mono_poly_order_carrier : mono_poly_order_carrier "\<lambda> x. True" "0 :: int" "op +" "op *" "op \<ge>" "op >" True 1 "max 0" int_name "\<lambda> x. True" "\<lambda> x. x > 0"
  by (unfold_locales, auto simp: Ring_and_Field.mult_strict_left_mono)

interpretation rat_mono_poly_order_carrier : mono_poly_order_carrier "\<lambda> x. True" "0 :: rat" "op +" "op *" "op \<ge>" "rat_gt \<delta>" True 1 "max 0" rat_name "\<lambda> x. True" "\<lambda> x. x \<ge> 1"
proof (unfold_locales, simp_all)
  fix x y z :: rat
  assume x: "1 \<le> x" and yz: "rat_gt \<delta> y z"
  have "\<exists> d > 0. rat_gt \<delta> = (\<lambda> x y. x - y \<ge> d)" 
  proof (cases "\<delta> > 0")
    case True
    show ?thesis 
      by (rule exI[of _ "min \<delta> 1"], simp add: True, intro ext, auto simp: rat_gt_def True)
  next
    case False
    show ?thesis 
      by (rule exI[of _ "1"], simp, intro ext, auto simp: rat_gt_def False)
  qed
  from this obtain d where d: "d > 0" and rat: "rat_gt \<delta> = (\<lambda> x y. x - y \<ge> d)" by auto
  from yz have yzd: "y - z \<ge> d" by (simp add: rat)
  show "rat_gt \<delta> (x * y) (x * z)"
  proof (simp only: rat)
    let ?p = "(x - 1) * (y - z)"
    from x have x1: "x - 1 \<ge> 0" by auto
    from yzd d have yz0: "y - z \<ge> 0" by auto
    have "?p \<ge> 0" 
      by (rule Ring_and_Field.pordered_cancel_semiring_class.mult_nonneg_nonneg[OF x1 yz0])
    have "x * y - x * z = x * (y - z)" using Ring_and_Field.ring_class.right_diff_distrib[of x y z] by auto
    also have "\<dots> = ((x - 1) + 1) * (y - z)" by auto
    also have "\<dots> = ?p + 1 * ( y - z)" by (rule Ring_and_Field.ring_class.ring_distribs(2))
    also have "\<dots> = ?p + (y - z)" by simp
    also have "\<dots> \<ge> 0 + d" using yzd `?p \<ge> 0` by auto
    finally 
    show "x * y - x * z \<ge> d" by auto
  qed
qed


interpretation int_mat_order: matrix_order "Suc n" "0 :: int" 1 "op +" "op *" "op \<ge>" "op >" "max 0" True "\<lambda> x :: int. True" 
  by (unfold_locales, auto simp: int_distrib Ring_and_Field.mult_right_mono Ring_and_Field.mult_left_mono, rule int_SN)

interpretation rat_mat_order: matrix_order "Suc n" "0 :: rat" 1 "op +" "op *" "op \<ge>" "rat_gt \<delta>" "max 0" True "\<lambda> x :: rat. True" 
  by (unfold_locales, auto simp: Ring_and_Field.mult_right_mono Ring_and_Field.mult_left_mono rat_poly_order_carrier.SN[simplified], auto simp: rat_gt_def)

interpretation arc_mat_order: matrix_order "Suc n" MinInfty "Num_arc 0" plus_arctic mult_arctic geq_arctic gt_arctic id False pos_arctic 
proof (unfold_locales, auto, 
    rule arctic_poly_carrier.plus_comm, simp+,
    rule arctic_poly_carrier.compat, simp+,
    rule arctic_poly_carrier.compat2, simp+,
    rule arctic_poly_carrier.plus_left_mono, simp+,
    rule arctic_poly_carrier.times_left_mono, simp+,
    rule arctic_poly_carrier.times_right_mono, simp+,
    rule arctic_poly_carrier.geq_trans, simp+,
    simp_all add: arctic_poly_carrier.zero_leastI
                  arctic_poly_carrier.zero_leastII 
                  arctic_poly_carrier.times_gt_left_mono
                  arctic_poly_carrier.plus_gt_both_mono) 
  from arctic_poly_carrier.zero_leastI arctic_poly_order_carrier.SN
  show "SN {(x, y). gt_arctic x y \<and> pos_arctic y}" by force
qed

interpretation int_mat_poly_order_carrier :  poly_order_carrier "wf_mat_rcI (Suc n) (Suc n)" "mat0I (0 :: int) (Suc n) (Suc n)" "mat_plusI (op + :: int \<Rightarrow> int \<Rightarrow> int)" "mat_mulI (0 :: int) (op +) (op *) (Suc n)" "mat_geI (op \<ge> :: int \<Rightarrow> int \<Rightarrow> bool)" "mat_gtI (op \<ge> :: int \<Rightarrow> int \<Rightarrow> bool) (op >)" True "mat1I (0 :: int) 1 (Suc n)" "mat_max0I (max (0 :: int))" int_mat_name "mat_arc_posI (\<lambda> x. True)"
  by (rule int_mat_order.mono_poly_order_carrier, auto)

interpretation rat_mat_poly_order_carrier :  poly_order_carrier "wf_mat_rcI (Suc n) (Suc n)" "mat0I (0 :: rat) (Suc n) (Suc n)" "mat_plusI (op + :: rat \<Rightarrow> rat \<Rightarrow> rat)" "mat_mulI (0 :: rat) (op +) (op *) (Suc n)" "mat_geI (op \<ge> :: rat \<Rightarrow> rat \<Rightarrow> bool)" "mat_gtI (op \<ge> :: rat \<Rightarrow> rat \<Rightarrow> bool) (rat_gt \<delta>)" True "mat1I (0 :: rat) 1 (Suc n)" "mat_max0I (max (0 :: rat))" rat_mat_name "mat_arc_posI (\<lambda> x. True)"
  by (rule rat_mat_order.mono_poly_order_carrier, auto)

interpretation arc_mat_poly_order_carrier : poly_order_carrier "wf_mat_rcI (Suc n) (Suc n)" "mat0I MinInfty (Suc n) (Suc n)" "mat_plusI plus_arctic" "mat_mulI MinInfty plus_arctic mult_arctic (Suc n)" "mat_geI geq_arctic" "mat_geI gt_arctic" False "mat1I MinInfty (Num_arc 0) (Suc n)" "mat_max0I id" arc_mat_name "mat_arc_posI pos_arctic"
  by (rule arc_mat_order.non_mono_poly_order_carrier, auto)

abbreviation nat_components :: "nat poly_components"
where "nat_components \<equiv> ((\<lambda> x. True) , 0 , 1 , op +, op *, id, op \<ge>, op >, True, nat_name, \<lambda> x. True)"

abbreviation int_components :: "int poly_components"
where "int_components \<equiv> ((\<lambda> x. True) , 0 , 1 , op +, op *, max 0, op \<ge>, op >, True, int_name, \<lambda> x. True)"

abbreviation int_mat_components :: "nat \<Rightarrow> int mat poly_components"
where "int_mat_components n \<equiv> (wf_mat_rcI (Suc n) (Suc n), mat0I 0 (Suc n) (Suc n) , mat1I 0 1 (Suc n), mat_plusI (op +) , mat_mulI 0 (op +) (op *) (Suc n), mat_max0I (max 0), mat_geI (op \<ge>), mat_gtI (op \<ge>) (op >), True, int_mat_name, mat_arc_posI (\<lambda> x. True))"

abbreviation rat_components :: "rat \<Rightarrow> rat poly_components"
where "rat_components \<delta> \<equiv> ((\<lambda> x. True) , 0 , 1 , op +, op *, max 0, op \<ge>, rat_gt \<delta>, True, rat_name, \<lambda> x. True)"

abbreviation rat_mat_components :: "rat \<Rightarrow> nat \<Rightarrow> rat mat poly_components"
where "rat_mat_components \<delta> n \<equiv> (wf_mat_rcI (Suc n) (Suc n), mat0I 0 (Suc n) (Suc n) , mat1I 0 1 (Suc n), mat_plusI (op +) , mat_mulI 0 (op +) (op *) (Suc n), mat_max0I (max 0), mat_geI (op \<ge>), mat_gtI (op \<ge>) (rat_gt \<delta>), True, rat_mat_name, mat_arc_posI (\<lambda> x. True))"

abbreviation arctic_components :: "arctic poly_components"
where "arctic_components \<equiv> ((\<lambda> x. True), MinInfty, Num_arc 0, plus_arctic, mult_arctic, id, geq_arctic, gt_arctic, False, arctic_name, pos_arctic)"

abbreviation arc_mat_components :: "nat \<Rightarrow> arctic mat poly_components"
where "arc_mat_components n \<equiv> (wf_mat_rcI (Suc n) (Suc n), mat0I MinInfty (Suc n) (Suc n) , mat1I MinInfty (Num_arc 0) (Suc n), mat_plusI plus_arctic, mat_mulI MinInfty plus_arctic mult_arctic (Suc n), mat_max0I id, mat_geI geq_arctic, mat_geI gt_arctic, False, arc_mat_name, mat_arc_posI pos_arctic)"

definition get_max_arity :: "('f,'v)trsL \<Rightarrow> nat"
where "get_max_arity R = foldr (\<lambda> n m. max n m) (map snd (concat (map (\<lambda> (l,r). funas_term_list l @ funas_term_list r) R))) 0"

fun lpo_repr_to_pr :: "'f lpo_repr \<Rightarrow> ('f filtered \<Rightarrow> nat \<Rightarrow> nat)"
where "lpo_repr_to_pr prs (FPair f a) ignored = (case lookup (f,a) prs of None \<Rightarrow> 0 | Some x \<Rightarrow> x)"

fun shows_afs :: "('f::shows)afs_list \<Rightarrow> shows"
where "shows_afs af = shows_map (\<lambda>((f,n),e). ''pi('' +#+ shows f +@+ ''/'' +#+ shows n +@+ '') = '' +#+ (case e of
  Collapse i \<Rightarrow> shows(Suc i)
| AFList ids \<Rightarrow> shows(map Suc ids)) +@+ shows_nl) af"
(*
fun afs_to_string :: "('f:: sshow) afs_list \<Rightarrow> string list"
where "afs_to_string af = map (\<lambda> ((f,n), e). ''pi( '' +++ sshow f +++ ''/'' +++ sshow n +++ '' ) = '' +++ (case e of Collapse i \<Rightarrow> sshow (Suc i) | AFList ids \<Rightarrow> show_list_gen ''[]'' ''['' '','' '']'' (map (\<lambda> i. sshow (Suc i)) ids))) af"
*)

fun shows_lpo_repr :: "('f::shows)lpo_repr \<Rightarrow> shows"
where "shows_lpo_repr prs = (''LPO with the following precedence'' +#+ shows_nl +@+ shows_map (\<lambda>((f,n),pr).
  ''precedence('' +#+ shows f +@+ ''['' +#+ shows n +@+ '']) = '' +#+ shows pr +@+ shows_nl
) prs +@+ shows_nl
+@+ ''precedence(_) = 0'' +#+ shows_nl
)"

(*
fun lpo_repr_to_string :: "('f :: sshow) lpo_repr \<Rightarrow> string list"
where "lpo_repr_to_string prs = ''LPO with the following precedence'' # map (\<lambda> ((f,n),pr). ''precedence('' +++ sshow f +++ ''['' +++ sshow n +++ '']) = '' +++ sshow pr) prs @ [''precedence(_) = 0'']"
*)

definition shows_lpo_af :: "('f::shows)lpo_repr \<Rightarrow> 'f afs_list \<Rightarrow> shows"
where "shows_lpo_af repr af \<equiv>
  shows_lpo_repr repr +@+ shows_nl +@+ ''in combination with the following argument filter'' +#+ shows_nl +@+
  shows_afs af
"

(*
definition lpo_af_to_string :: "('f :: sshow) lpo_repr \<Rightarrow> 'f afs_list \<Rightarrow> string list"
where "lpo_af_to_string repr af \<equiv> lpo_repr_to_string repr @ ([] # ''in combination with the following argument filter'' #  afs_to_string af)"
*)

definition lpo_repr where "lpo_repr \<equiv> shows_string ''LPO+AF''"
definition lpoS where "lpoS \<equiv> shows_string ''>_LPO+AF''"
definition lpoNS where "lpoNS \<equiv> shows_string ''>=_LPO+AF''"

fun create_rule_removal_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow>
        shows monad"
where "create_rule_removal_proc (Int_carrier I) = create_rule_removal_proc_comp int_components (\<lambda> x. x > 0) I"
  |   "create_rule_removal_proc (Nat_carrier I) = create_rule_removal_proc_comp nat_components (\<lambda> x. x \<noteq> 0) I"
  |   "create_rule_removal_proc (Rat_carrier \<delta> I) = create_rule_removal_proc_comp (rat_components \<delta>) (\<lambda> x. x \<ge> 1) I"
  |   "create_rule_removal_proc (Arctic_carrier I) = (\<lambda> _ _. error(''arctic not monotonic for TRSs'' +#+ shows_nl))"
  |   "create_rule_removal_proc (Arc_mat_carrier n I) = (\<lambda> _ _. error(''arctic not monotonic for TRSs'' +#+ shows_nl))"
  |   "create_rule_removal_proc (Int_mat_carrier n I) = (\<lambda> _ _. error(''monotonic matrix-orders not yet supported'' +#+ shows_nl))"
  |   "create_rule_removal_proc (Rat_mat_carrier \<delta> n I) = (\<lambda> _ _. error(''monotonic matrix-orders not yet supported'' +#+ shows_nl))"
  |   "create_rule_removal_proc (LPO pri pi) = (\<lambda> R rR. let m = get_max_arity R; pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in check_wf_afs pi \<guillemotright> checkRuleRemovalProc (check_mono_afs pi) (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) lpo_repr (shows_lpo_af pri pi) R rR)" 

fun create_redpair_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_redpair_proc (Int_carrier I) = create_redpair_proc_comp int_components I"
  |   "create_redpair_proc (Nat_carrier I) = create_redpair_proc_comp nat_components I"
  |   "create_redpair_proc (Rat_carrier \<delta> I) = create_redpair_proc_comp (rat_components \<delta>) I"
  |   "create_redpair_proc (Arctic_carrier I) = create_redpair_proc_comp arctic_components I"
  |   "create_redpair_proc (Int_mat_carrier n I) = create_redpair_proc_comp (int_mat_components n) I"
  |   "create_redpair_proc (Rat_mat_carrier \<delta> n I) = create_redpair_proc_comp (rat_mat_components \<delta> n) I"
  |   "create_redpair_proc (Arc_mat_carrier n I) = create_redpair_proc_comp (arc_mat_components n) I"
  |   "create_redpair_proc (LPO pri pi) = (\<lambda> P rP R. let m = max (get_max_arity R) (get_max_arity P); pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in check_wf_afs pi \<guillemotright> checkRedPairProc (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) lpo_repr (shows_lpo_af pri pi) P rP R)"

fun create_mono_redpair_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_mono_redpair_proc (Int_carrier I) = create_mono_redpair_proc_comp int_components (\<lambda> x. x > 0) I"
  |   "create_mono_redpair_proc (Nat_carrier I) = create_mono_redpair_proc_comp nat_components (\<lambda> x. x \<noteq> 0) I"
  |   "create_mono_redpair_proc (Rat_carrier \<delta> I) = create_mono_redpair_proc_comp (rat_components \<delta>) (\<lambda> x. x \<ge> 1) I"
  |   "create_mono_redpair_proc (Arctic_carrier I) = (\<lambda> _ _ _ _. error(''arctic not monotonic for TRSs'' +#+ shows_nl))"
  |   "create_mono_redpair_proc (Arc_mat_carrier n I) = (\<lambda> _ _ _ _. error(''arctic not monotonic for TRSs'' +#+ shows_nl))"
  |   "create_mono_redpair_proc (Int_mat_carrier n I) = (\<lambda> _ _ _ _. error(''monotonic matrix-orders not yet supported'' +#+ shows_nl))"
  |   "create_mono_redpair_proc (Rat_mat_carrier _ _ I) = (\<lambda> _ _ _ _. error(''monotonic matrix-orders not yet supported'' +#+ shows_nl))"
  |   "create_mono_redpair_proc (LPO pri pi) = (\<lambda> P rP R rR. let m = max (get_max_arity R) (get_max_arity P); pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in check_wf_afs pi \<guillemotright> checkMonoRedPairProc (check_mono_afs pi) (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) lpo_repr (shows_lpo_af pri pi) P rP R rR)"


fun create_redpair_ur_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_redpair_ur_proc (Int_carrier I) = create_redpair_ur_proc_comp int_components I"
  |   "create_redpair_ur_proc (Nat_carrier I) = create_redpair_ur_proc_comp nat_components I"
  |   "create_redpair_ur_proc (Rat_carrier \<delta> I) = create_redpair_ur_proc_comp (rat_components \<delta>) I"
  |   "create_redpair_ur_proc (Arctic_carrier I) = create_redpair_ur_proc_comp arctic_components I"
  |   "create_redpair_ur_proc (Int_mat_carrier n I) = create_redpair_ur_proc_comp (int_mat_components n) I"
  |   "create_redpair_ur_proc (Rat_mat_carrier \<delta> n I) = create_redpair_ur_proc_comp (rat_mat_components \<delta> n) I"
  |   "create_redpair_ur_proc (Arc_mat_carrier n I) = create_redpair_ur_proc_comp (arc_mat_components n) I"
  |   "create_redpair_ur_proc (LPO pri pi) = (\<lambda> P rP R rm ur. let m = max (get_max_arity R) (get_max_arity P); pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in 
                check_wf_afs pi \<guillemotright> checkRedPairUrProc (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) lpo_repr (shows_lpo_af pri pi) P rP R rm ur)"

fun create_mono_redpair_ur_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_mono_redpair_ur_proc (Int_carrier I) = create_mono_redpair_ur_proc_comp int_components (\<lambda> x. x > 0) I"
  |   "create_mono_redpair_ur_proc (Nat_carrier I) = create_mono_redpair_ur_proc_comp nat_components (\<lambda> x. x \<noteq> 0) I"
  |   "create_mono_redpair_ur_proc (Rat_carrier \<delta> I) = create_mono_redpair_ur_proc_comp (rat_components \<delta>) (\<lambda> x. x \<ge> 1) I"
  |   "create_mono_redpair_ur_proc (Arctic_carrier I) = (\<lambda> _ _ _ _ _ _. error(''arctic not monotonic for TRSs'' +#+ shows_nl))"
  |   "create_mono_redpair_ur_proc (Arc_mat_carrier n I) = (\<lambda> _ _ _ _ _ _. error(''arctic not monotonic for TRSs'' +#+ shows_nl))"
  |   "create_mono_redpair_ur_proc (Int_mat_carrier n I) = (\<lambda> _ _ _ _ _ _. error(''monotonic matrix-orders not yet supported'' +#+ shows_nl))"
  |   "create_mono_redpair_ur_proc (Rat_mat_carrier _ n I) = (\<lambda> _ _ _ _ _ _. error(''monotonic matrix-orders not yet supported'' +#+ shows_nl))"
  |   "create_mono_redpair_ur_proc (LPO pri pi) = (\<lambda> P rP R rR rm ur. let m = max (get_max_arity R) (get_max_arity P); pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in 
                check_wf_afs pi \<guillemotright> checkMonoRedPairUrProc (check_mono_afs pi) (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) lpo_repr (shows_lpo_af pri pi) P rP R rR rm ur)" 


fun create_redpair_ur_af_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        shows monad"
where "create_redpair_ur_af_proc (Int_carrier I) = create_redpair_ur_af_proc_comp int_components I"
  |   "create_redpair_ur_af_proc (Nat_carrier I) = create_redpair_ur_af_proc_comp nat_components I"
  |   "create_redpair_ur_af_proc (Rat_carrier \<delta> I) = create_redpair_ur_af_proc_comp (rat_components \<delta>) I"
  |   "create_redpair_ur_af_proc (Arctic_carrier I) = create_redpair_ur_af_proc_comp arctic_components I"
  |   "create_redpair_ur_af_proc (Int_mat_carrier n I) = create_redpair_ur_af_proc_comp (int_mat_components n) I"
  |   "create_redpair_ur_af_proc (Rat_mat_carrier \<delta> n I) = create_redpair_ur_af_proc_comp (rat_mat_components \<delta> n) I"
  |   "create_redpair_ur_af_proc (Arc_mat_carrier n I) = create_redpair_ur_af_proc_comp (arc_mat_components n) I"
  |   "create_redpair_ur_af_proc (LPO pri pi) = (\<lambda> P rP R rm ur. let m = max (get_max_arity R) (get_max_arity P); pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in 
                check_wf_afs pi \<guillemotright> checkRedPairUrAfProc (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) lpo_repr (shows_lpo_af pri pi) P rP R rm ur (afs_list_to_af pi))"

fun create_sct_ur_af_proc :: "('f::shows)redpair \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> (('f,'v)rule \<times> ((nat \<times> nat)list \<times> (nat \<times> nat)list))list \<Rightarrow>
        shows monad" 
where "create_sct_ur_af_proc (Int_carrier I) = create_sct_ur_af_proc_comp int_components I"
  |   "create_sct_ur_af_proc (Nat_carrier I) = create_sct_ur_af_proc_comp nat_components I"
  |   "create_sct_ur_af_proc (Rat_carrier \<delta> I) = create_sct_ur_af_proc_comp (rat_components \<delta>) I"
  |   "create_sct_ur_af_proc (Arctic_carrier I) = create_sct_ur_af_proc_comp arctic_components I"
  |   "create_sct_ur_af_proc (Int_mat_carrier n I) = create_sct_ur_af_proc_comp (int_mat_components n) I"
  |   "create_sct_ur_af_proc (Rat_mat_carrier \<delta> n I) = create_sct_ur_af_proc_comp (rat_mat_components \<delta> n) I"
  |   "create_sct_ur_af_proc (Arc_mat_carrier n I) = create_sct_ur_af_proc_comp (arc_mat_components n) I"
  |   "create_sct_ur_af_proc (LPO pri pi) = (\<lambda> P R rm rrm ur graphs. let m = max (get_max_arity R) (get_max_arity P); pr = lpo_repr_to_pr pri; \<pi> = afs_list_to_afs pi in 
                check_wf_afs pi \<guillemotright> check_sct_ur_af_proc (afs_check lpoS \<pi> (lpo_strict pr m)) (afs_check lpoNS \<pi> (lpo_nstrict pr m)) (shows_lpo_af pri pi) P graphs R rm rrm ur (afs_list_to_af pi))"


lemma create_rule_removal_proc: 
  shows "ci_tt_impl (\<lambda> R rR. isOK(create_rule_removal_proc I R rR))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> R rR. isOK(create_rule_removal_proc I R rR)) = 
    (\<lambda> R rR. isOK(create_rule_removal_proc_comp nat_components (\<lambda> x. x \<noteq> 0) nI R rR))" by auto
  show ?thesis by (simp only: id, rule nat_mono_poly_order_carrier.rule_removal_tt_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> R rR. isOK(create_rule_removal_proc I R rR)) = 
    (\<lambda> R rR. isOK(create_rule_removal_proc_comp int_components (\<lambda> x. x > 0) nI R rR))" by auto
  show ?thesis by (simp only: id, rule int_mono_poly_order_carrier.rule_removal_tt_ci)
next
  case (Rat_carrier d nI)
  hence id: "(\<lambda> R rR. isOK(create_rule_removal_proc I R rR)) = 
    (\<lambda> R rR. isOK(create_rule_removal_proc_comp (rat_components d) (\<lambda> x. x \<ge> 1) nI R rR))" by auto
  show ?thesis by (simp only: id, rule rat_mono_poly_order_carrier.rule_removal_tt_ci[of d])
next
  case (Arctic_carrier nI) thus ?thesis unfolding ci_tt_impl_def by auto
next
  case (Int_mat_carrier n nI) thus ?thesis unfolding ci_tt_impl_def by auto
next
  case (Rat_mat_carrier d n nI) thus ?thesis unfolding ci_tt_impl_def by auto
next
  case (Arc_mat_carrier n nI) thus ?thesis unfolding ci_tt_impl_def by auto
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> R. get_max_arity R"
  let ?pi = "afs_list_to_afs pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> R rR. isOK(create_rule_removal_proc I R rR)) = 
    (\<lambda> R rR. isOK(check_wf_afs pi \<guillemotright> checkRuleRemovalProc (OK \<guillemotright> check_mono_afs pi) (?chS (lpo_strict ?pr (?m R))) (?ch (lpo_nstrict ?pr (?m R))) lpo_repr ?pres R rR))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold ci_tt_impl_def, intro allI impI)
    fix R rR
    let ?check1 = "check_wf_afs pi"
    let ?check2 = "checkRuleRemovalProc (OK \<guillemotright> check_mono_afs pi) (?chS (lpo_strict ?pr (?m R))) (?ch (lpo_nstrict ?pr (?m R))) lpo_repr ?pres R rR"
    assume check: "isOK(?check1 \<guillemotright> ?check2)"
    from check have c1: "isOK(?check1)" and c2: "isOK(?check2)" by auto
    from c2 LPO_afs_impl.rule_removal[unfolded LPO_afs_impl_def, of pi, OF c1]
    show "set rR \<subseteq>ci set R \<and> (\<forall> t. ideriv (set R) t \<longrightarrow> (\<exists> i. ideriv (set rR) (shift t i)))"
      unfolding ci_tt_impl_def by blast
  qed
qed

lemma create_redpair_proc: 
  fixes I :: "('f::shows)redpair"
  shows "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp nat_components nI P rP R))" by auto
  show ?thesis by (simp only: id, rule nat_poly_order_carrier.redpair_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp int_components nI P rP R))" by auto
  show ?thesis by (simp only: id,  rule int_poly_order_carrier.redpair_proc_comp_ci)
next
  case (Rat_carrier d nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp (rat_components d) nI P rP R))" by auto
  show ?thesis by (simp only: id,  rule rat_poly_order_carrier.redpair_proc_comp_ci)
next
  case (Arctic_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp arctic_components nI P rP R))" by auto
  show ?thesis by (simp only: id,  rule arctic_poly_order_carrier.redpair_proc_comp_ci)
next
  case (Int_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp (int_mat_components n) nI P rP R))" by auto
  show ?thesis by (simp only: id, rule int_mat_poly_order_carrier.redpair_proc_comp_ci)
next
  case (Rat_mat_carrier d n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp (rat_mat_components d n) nI P rP R))" by auto
  show ?thesis by (simp only: id, rule rat_mat_poly_order_carrier.redpair_proc_comp_ci)
next
  case (Arc_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp (arc_mat_components n) nI P rP R))" by auto
  show ?thesis by (simp only: id, rule arc_mat_poly_order_carrier.redpair_proc_comp_ci)
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> P R. max (get_max_arity R) (get_max_arity P)"
  let ?pi = "afs_list_to_afs pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(check_wf_afs pi \<guillemotright> checkRedPairProc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold chain_identifying_onlyDP_proc_impl_def, intro allI impI)
    fix P R rP :: "('f,'v::shows)trsL" and s t \<sigma>
    let ?check1 = "check_wf_afs pi"
    let ?check2 = "checkRedPairProc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R"
    assume check: "isOK(?check1 \<guillemotright> ?check2)"
      and chain: "min_ichain (set P, set R) s t \<sigma>"
    from check have c1: "isOK(?check1)" and c2: "isOK(?check2)" by auto
    from chain c2 LPO_afs_impl.redpair_proc[unfolded LPO_afs_impl_def, of pi, OF c1]
    show "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_onlyDP_proc_impl_def by blast
  qed
qed

lemma create_mono_redpair_proc: "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp nat_components (\<lambda> x. x \<noteq> 0) nI P rP R rR))" by auto
  show ?thesis by (simp only: id, rule nat_mono_poly_order_carrier.mono_redpair_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp int_components (\<lambda> x. x > 0) nI P rP R rR))" by auto
  show ?thesis by (simp only: id, rule int_mono_poly_order_carrier.mono_redpair_proc_comp_ci)
next
  case (Rat_carrier d nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp (rat_components d) (\<lambda> x. x \<ge> 1) nI P rP R rR))" by auto
  show ?thesis by (simp only: id, rule rat_mono_poly_order_carrier.mono_redpair_proc_comp_ci)
next
  case (Arctic_carrier nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (Int_mat_carrier n nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (Rat_mat_carrier d n nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (Arc_mat_carrier n nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> P R. max (get_max_arity R) (get_max_arity P)"
  let ?pi = "afs_list_to_afs pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(check_wf_afs pi \<guillemotright> checkMonoRedPairProc (OK \<guillemotright> check_mono_afs pi) (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R rR))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold chain_identifying_proc_impl_def, intro allI impI)
    fix P R rP rR 
    let ?check1 = "check_wf_afs pi"
    let ?check2 = "checkMonoRedPairProc (OK \<guillemotright> check_mono_afs pi) (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R rR"
    assume check: "isOK(?check1 \<guillemotright> ?check2)"
    from check have c1: "isOK(?check1)" and c2: "isOK(?check2)" by auto
    from c2 LPO_afs_impl.mono_redpair_proc[unfolded LPO_afs_impl_def, of pi, OF c1]
    show "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))"
      unfolding chain_identifying_proc_impl_def by blast
  qed
qed

lemma create_redpair_ur_proc: 
  fixes I :: "('f::shows)redpair" and ur :: "('f,'v::shows)trsL"
  shows "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp nat_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule nat_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp int_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule int_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (Rat_carrier d nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp (rat_components d) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule rat_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (Arctic_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp arctic_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule arctic_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (Int_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp (int_mat_components n) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule int_mat_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (Rat_mat_carrier d n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp (rat_mat_components d n) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule rat_mat_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (Arc_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp (arc_mat_components n) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule arc_mat_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> P R. max (get_max_arity R) (get_max_arity P)"
  let ?pi = "afs_list_to_afs pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(check_wf_afs pi \<guillemotright> checkRedPairUrProc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R rm ur))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold chain_identifying_onlyDP_proc_impl_def, intro allI impI)
    fix P R rP :: "('f::shows,'v)trsL" and s t \<sigma>
    let ?c1 = "check_wf_afs pi"
    let ?c2 = "checkRedPairUrProc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R (computeRuleMap R) ur"
    assume check: "isOK(?c1 \<guillemotright> ?c2)"
      and min: "min_ichain (set P, set R) s t \<sigma>"
    from check have c1: "isOK(?c1)" and c2: "isOK(?c2)" by auto
    from min c2 LPO_afs_impl.redpair_ur_proc[unfolded LPO_afs_impl_def, OF c1]
    show "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_onlyDP_proc_impl_def by blast
  qed
qed


lemma create_redpair_ur_af_proc: fixes ur :: "('f::shows,'v::shows)trsL"
  shows "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp nat_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule nat_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next 
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp int_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule int_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next
  case (Rat_carrier d nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp (rat_components d) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule rat_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next
  case (Arctic_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp arctic_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule arctic_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next
  case (Int_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp (int_mat_components n) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule int_mat_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next
  case (Rat_mat_carrier d n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp (rat_mat_components d n) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule rat_mat_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next
  case (Arc_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp (arc_mat_components n) nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule arc_mat_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> P R. max (get_max_arity R) (get_max_arity P)"
  let ?pi = "afs_list_to_afs pi"
  let ?af = "afs_list_to_af pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(check_wf_afs pi \<guillemotright> checkRedPairUrAfProc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R rm ur ?af))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold chain_identifying_onlyDP_proc_impl_def, intro allI impI)
    fix P R rP :: "('f::shows,'v)trsL" and s t \<sigma>
    let ?c1 = "check_wf_afs pi"
    let ?c2 = "checkRedPairUrAfProc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R (computeRuleMap R) ur ?af"
    assume check: "isOK(?c1 \<guillemotright> ?c2)"
      and min: "min_ichain (set P, set R) s t \<sigma>"
    from check have c1: "isOK(?c1)" and c2: "isOK(?c2)" by auto
    from min c2 LPO_afs_impl.redpair_ur_af_proc[unfolded LPO_afs_impl_def, OF c1]
    show "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_onlyDP_proc_impl_def by blast
  qed
qed

lemma create_mono_redpair_ur_proc: "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp nat_components (\<lambda> x. x \<noteq> 0) nI P rP R rm rR ur))" by auto
  show ?thesis by (simp only: id, rule nat_mono_poly_order_carrier.mono_redpair_ur_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp int_components (\<lambda> x. x > 0) nI P rP R rm rR ur))" by auto
  show ?thesis by (simp only: id, rule int_mono_poly_order_carrier.mono_redpair_ur_proc_comp_ci)
next
  case (Rat_carrier d nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp (rat_components d) (\<lambda> x. x \<ge> 1) nI P rP R rm rR ur))" by auto
  show ?thesis by (simp only: id, rule rat_mono_poly_order_carrier.mono_redpair_ur_proc_comp_ci)
next
  case (Arctic_carrier nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (Int_mat_carrier nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (Rat_mat_carrier d nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (Arc_mat_carrier nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> P R. max (get_max_arity R) (get_max_arity P)"
  let ?pi = "afs_list_to_afs pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(check_wf_afs pi \<guillemotright> checkMonoRedPairUrProc (OK \<guillemotright> check_mono_afs pi) (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R rm rR ur))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold chain_identifying_proc_impl_def, intro allI impI)
    fix P R rP rR 
    let ?check1 = "check_wf_afs pi"
    let ?check2 = "checkMonoRedPairUrProc (OK \<guillemotright> check_mono_afs pi) (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) lpo_repr ?pres P rP R (computeRuleMap R) rR ur"
    assume check: "isOK(?check1 \<guillemotright> ?check2)"
    from check have c1: "isOK(?check1)" and c2: "isOK(?check2)" by auto
    from c2 LPO_afs_impl.mono_redpair_ur_proc[unfolded LPO_afs_impl_def, of pi, OF c1]
    show "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))"
      unfolding chain_identifying_proc_impl_def by blast
  qed
qed

context LPO_afs_impl
begin
lemmas sct_proc =  check_sct_ur_af_proc_sound
end

lemma create_sct_ur_af_proc: fixes ur :: "('f::shows,'v::shows)trsL"
  shows "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P R rm rrm ur graphs))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P R rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp nat_components nI P R rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id,  rule nat_poly_order_carrier.sct_ur_af_proc_comp_ci)
next 
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P R rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp int_components nI P R rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id,  rule int_poly_order_carrier.sct_ur_af_proc_comp_ci)
next 
  case (Rat_carrier d nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P R rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp (rat_components d) nI P R rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id,  rule rat_poly_order_carrier.sct_ur_af_proc_comp_ci)
next 
  case (Arctic_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P R rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp arctic_components nI P R rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id,  rule arctic_poly_order_carrier.sct_ur_af_proc_comp_ci)
next
  case (Int_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P R rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp (int_mat_components n) nI P R rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id, rule int_mat_poly_order_carrier.sct_ur_af_proc_comp_ci)
next
  case (Rat_mat_carrier d n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P R rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp (rat_mat_components d n) nI P R rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id, rule rat_mat_poly_order_carrier.sct_ur_af_proc_comp_ci)
next
  case (Arc_mat_carrier n nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P R rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp (arc_mat_components n) nI P R rm rrm ur graphs))" by auto
  show ?thesis by (simp only: id, rule arc_mat_poly_order_carrier.sct_ur_af_proc_comp_ci)
next
  case (LPO pri pi)
  let ?pr = "lpo_repr_to_pr pri"
  let ?pres = "shows_lpo_af pri pi"
  let ?m = "\<lambda> P R. max (get_max_arity R) (get_max_arity P)"
  let ?pi = "afs_list_to_afs pi"
  let ?af = "afs_list_to_af pi"
  let ?chS = "afs_check lpoS ?pi"
  let ?ch = "afs_check lpoNS ?pi"
  from LPO have id: "(\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc I P R rm rrm ur graphs)) = 
    (\<lambda> P R rm rrm rP. isOK(check_wf_afs pi \<guillemotright> check_sct_ur_af_proc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) ?pres P graphs R rm rrm ur ?af))" by (simp add: Let_def)
  show ?thesis 
  proof (simp only: id, unfold chain_identifying_onlyDP_proc_impl_def, intro allI impI)
    fix P R rP s t \<sigma>
    let ?c1 = "check_wf_afs pi"
    let ?c2 = "check_sct_ur_af_proc (?chS (lpo_strict ?pr (?m P R))) (?ch (lpo_nstrict ?pr (?m P R))) ?pres P graphs R (computeRuleMap R) (computeRuleMap (reverse_trs R)) ur ?af"
    assume check: "isOK(?c1 \<guillemotright> ?c2)"
      and min: "min_ichain (set P, set R) s t \<sigma>"
    hence c1: "isOK(?c1)" and c2: "isOK(?c2)" by auto
    from min c2 LPO_afs_impl.sct_proc[unfolded LPO_afs_impl_def, OF c1]
    show "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_onlyDP_proc_impl_def by blast
  qed
qed
  
end